import re
import textwrap
from functools import partial
from importlib import import_module
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple

from docutils.utils import column_width
from mistune import Markdown
from mistune.core import BaseRenderer, BlockState
from mistune.plugins import _plugins

from .parse import RestBlockParser, RestInlineParser

CACHED_MODULES: Dict[str, Any] = {}
DEFAULT_PLUGINS = ["strikethrough", "footnotes", "table"]

PROLOG = """\
.. role:: raw-html-md(raw)
   :format: html

"""


class RestRenderer(BaseRenderer):
    _include_raw_html = False
    indent = " " * 3
    list_marker = "{#__rest_list_mark__#}"
    hmarks = {
        1: "=",
        2: "-",
        3: "^",
        4: "~",
        5: '"',
        6: "#",
    }

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self._indent_block = partial(textwrap.indent, prefix=self.indent)
        super().__init__(*args, **kwargs)

    def render_token(self, token: Dict[str, Any], state: BlockState) -> str:
        # based on mistune 3.0.2, mistune/renderers/html.py
        func: Callable[..., str] = self._get_method(token["type"])
        attrs = token.get("attrs")
        style = token.get("style")

        if "raw" in token:
            text = token["raw"]
        elif "children" in token:
            text = self.render_tokens(token["children"], state)
        else:
            if attrs:
                return func(**attrs)
            else:
                return func()

        # We have to special-case block_code, as it needs to know the
        # style as well to determine whether to add a blank line at the
        # end (so as to retain the original behaviour)
        if token["type"] == "block_code":
            if attrs:
                return func(text, style=style, **attrs)
            else:
                return func(text, style=style)

        if attrs:
            return func(text, **attrs)
        else:
            return func(text)

    def finalize(self, data: Iterable[str]) -> str:
        return "".join(data)

    def _raw_html(self, html: str) -> str:
        self._include_raw_html = True
        return r":raw-html-md:`{}`".format(html)

    def block_code(self, code: str, style: str, info: Optional[str] = None) -> str:
        if info == "math":
            first_line = "\n.. math::\n\n"
        elif info:
            first_line = "\n.. code-block:: {}\n\n".format(info)
        else:
            # first_line = "\n::\n\n"
            first_line = "\n.. code-block::\n\n"
        newline = "\n" if style == "indent" else ""
        return first_line + self._indent_block(code + newline)

    def block_quote(self, text: str) -> str:
        # text includes some empty line
        return "\n..\n\n{}\n\n".format(self._indent_block(text.strip("\n")))

    def block_text(self, text: str) -> str:
        return text

    def block_html(self, html: str) -> str:
        """Rendering block level pure html content.

        :param html: text content of the html snippet.
        """
        return "\n\n.. raw:: html\n\n" + self._indent_block(html) + "\n"

    def heading(self, text: str, level: int, **attrs: Any) -> str:
        """Rendering header/heading tags like ``<h1>`` ``<h2>``.

        :param text: rendered text content for the header.
        :param level: a number for the header level, for example: 1.
        :param attrs: other attributes of the header.
        """
        return "\n{0}\n{1}\n".format(text, self.hmarks[level] * column_width(text))

    def thematic_break(self) -> str:
        """Rendering method for ``<hr>`` tag."""
        return "\n----\n"

    def list(self, text: str, ordered: bool, **attrs: Any) -> str:
        """Rendering list tags like ``<ul>`` and ``<ol>``.

        :param text: body contents of the list.
        :param ordered: whether this list is ordered or not.
        :param attrs: other attributes of the list.
        """
        mark = "#. " if ordered else "* "
        lines = text.splitlines()
        for i, line in enumerate(lines):
            if line and not line.startswith(self.list_marker):
                lines[i] = " " * len(mark) + line
        result = "\n{}\n".format("\n".join(lines)).replace(self.list_marker, mark)
        return result

    def list_item(self, text: str) -> str:
        """Rendering list item snippet. Like ``<li>``."""
        return "\n" + self.list_marker + text

    def paragraph(self, text: str) -> str:
        """Rendering paragraph tags. Like ``<p>``."""
        return "\n" + text + "\n"

    def table(self, body: str) -> str:
        """Rendering table element. Wrap header and body in it.

        :param header: header part of the table.
        :param body: body part of the table.
        """
        table = "\n.. list-table::\n"
        table = table + self._indent_block(body) + "\n"
        return table

    def table_head(self, text: str) -> str:
        return ":header-rows: 1\n\n" + self.table_row(text)

    def table_body(self, text: str) -> str:
        return text

    def table_row(self, content: str) -> str:
        """Rendering a table row. Like ``<tr>``.

        :param content: content of current table row.
        """
        contents = content.splitlines()
        if not contents:
            return ""
        clist = ["* " + contents[0]]
        if len(contents) > 1:
            for c in contents[1:]:
                clist.append("  " + c)
        return "\n".join(clist) + "\n"

    def table_cell(self, content: str, align: None = None, head: bool = False) -> str:
        """Rendering a table cell. Like ``<th>`` ``<td>``.

        :param content: content of current table cell.
        :param align: align of current table cell.
        :param head: whether this is header or not.
        """
        return "- " + content + "\n"

    def double_emphasis(self, text: str) -> str:
        """Rendering **strong** text.

        :param text: text content for emphasis.
        """
        return r"**{}**".format(text)

    def emphasis(self, text: str) -> str:
        """Rendering *emphasis* text.

        :param text: text content for emphasis.
        """
        return r"*{}*".format(text)

    def strong(self, text: str) -> str:
        return r"**{}**".format(text)

    def codespan(self, text: str) -> str:
        """Rendering inline `code` text.

        :param text: text content for inline code.
        """
        cannot_inline = "``" in text or text[0] in [" ", "`"] or text[-1] in [" ", "`"]
        if cannot_inline:
            # actually, docutils split spaces in literal
            return self._raw_html(
                '<code class="docutils literal">'
                '<span class="pre">{}</span>'
                "</code>".format(text.replace("`", "&#96;"))
            )
        else:
            return r"``{}``".format(text)

    def linebreak(self) -> str:
        """Rendering line break like ``<br>``."""
        return " " + self._raw_html("<br />") + "\n"

    def softbreak(self) -> str:
        """Rendering soft line break."""
        return "\n"

    def strikethrough(self, text: str) -> str:
        """Rendering ~~strikethrough~~ text.

        :param text: text content for strikethrough.
        """
        return self._raw_html("<del>{}</del>".format(text))

    def text(self, text: str) -> str:
        """Rendering unformatted text.

        :param text: text content.
        """
        return text

    def link(self, text: str, url: str, title: Optional[str] = None) -> str:
        """Rendering a given link with content and title.

        :param text: text content for description.
        :param url: URL for ``<a>`` tag.
        :param title: title content for `title` attribute.
        """
        if text.startswith("\n.. image::"):
            text = re.sub(r":target: (.*)\n", f":target: {url}\n", text)
            return text

        if text.startswith("``") and text.endswith("``"):
            # Return raw HTML for inline code:
            html = (
                '<code class="docutils literal">'
                '<span class="pre">{}</span>'
                "</code>".format(text[2:-2].replace("`", "&#96;"))
            )
            return self._raw_html(
                '<a href="{url}">{text}</a>'.format(url=url, text=html)
            )

        underscore = "_"
        if title:
            return self._raw_html(
                '<a href="{url}" title="{title}">{text}</a>'.format(
                    url=url, title=title, text=text
                )
            )
        if url.startswith("#"):
            target = url[1:]
            return r":ref:`{text} <{target}>`".format(target=target, text=text)

        return r"`{text} <{target}>`{underscore}".format(
            target=url, text=text, underscore=underscore
        )

    def image(self, text: str, url: str, title: Optional[str] = None) -> str:
        """Rendering a image with title and text.

        :param text: alt text of the image.
        :param url: source link of the image.
        :param title: title text of the image.
        """
        # rst does not support title option
        # and I couldn't find title attribute in HTML standard
        return "\n".join(
            [
                "",
                ".. image:: {}".format(url),
                "   :target: {}".format(url),
                "   :alt: {}".format(text),
                "",
            ]
        )

    def image_link(self, url: str, target: str, alt: str) -> str:
        return "\n".join(
            [
                "",
                ".. image:: {}".format(url),
                "   :target: {}".format(target),
                "   :alt: {}".format(alt),
                "",
            ]
        )

    def inline_html(self, html: str) -> str:
        """Rendering span level pure html content.

        :param html: text content of the html snippet.
        """
        return self._raw_html(html)

    def newline(self) -> str:
        """Rendering newline element."""
        return ""

    def footnote_ref(self, key: str, index: int) -> str:
        """Rendering the ref anchor of a footnote.

        :param key: identity key for the footnote.
        :param index: the index count of current footnote.
        """
        return r"[#fn-{}]_".format(key)

    def footnote_item(self, text: str, key: str, index: int) -> str:
        """Rendering a footnote item.

        :param key: identity key for the footnote.
        :param text: text content of the footnote.
        """
        return ".. [#fn-{0}] {1}\n".format(key, text.strip())

    def footnotes(self, text: str) -> str:
        """Wrapper for all footnotes.

        :param text: contents of all footnotes.
        """
        if text:
            return "\n\n" + text
        else:
            return ""

    """Below outputs are for rst."""

    def rest_role(self, raw: str) -> str:
        return raw

    def rest_link(self, raw: str) -> str:
        return raw

    def inline_math(self, raw: str) -> str:
        """Extension of recommonmark."""
        return r":math:`{}`".format(raw)

    def eol_literal_marker(self, raw: str) -> str:
        """Extension of recommonmark."""
        return raw

    def directive(self, text: str) -> str:
        return "\n" + text

    def rest_code_block(self, text: str) -> str:
        return "\n\n"

    def blank_line(self) -> str:
        return ""


class RestMarkdown(Markdown):
    def __init__(
        self,
        renderer: Optional[BaseRenderer] = None,
        block: Optional[RestBlockParser] = None,
        inline: Optional[RestInlineParser] = None,
        plugins: Optional[List[Any]] = None,
        **kwargs: Any,
    ) -> None:
        renderer = renderer or RestRenderer()
        block = block or RestBlockParser()
        inline = inline or RestInlineParser()
        plugins_str = plugins or [_plugins[p] for p in DEFAULT_PLUGINS]
        plugins = []
        for plugin_str in plugins_str:
            if plugin_str in CACHED_MODULES:
                plugins.append(CACHED_MODULES[plugin_str])
            else:
                if isinstance(plugin_str, str):
                    module_path, func_name = plugin_str.rsplit(".", 1)
                    module = import_module(module_path)
                    plugin = getattr(module, func_name)
                else:
                    # Presumably a function has been passed
                    plugin = plugin_str
                CACHED_MODULES[plugin_str] = plugin
                plugins.append(plugin)

        super().__init__(renderer, block=block, inline=inline, plugins=plugins)

    def parse(
        self,
        text: str,
        state: Optional[BlockState] = None,
    ) -> Tuple[str, Optional[BlockState]]:
        output, state = super().parse(text)
        output = self.post_process(output)

        return output, state

    def post_process(self, text: str) -> str:
        if self.renderer._include_raw_html:
            return PROLOG + text
        else:
            return text


def convert(text: str, **kwargs: Any) -> str:
    return str(RestMarkdown(**kwargs)(text))



# Copyright Amethyst Reese
# Licensed under the MIT License

import platform
import unittest
from dataclasses import dataclass, field
from pathlib import Path
from textwrap import dedent

from docutils.frontend import get_default_settings
from docutils.parsers.rst import directives, Parser
from docutils.utils import new_document

from ..render import convert
from ..sphinx import MdInclude

TEST_MD = Path(__file__).parent / "test.md"
TEST_RST = Path(__file__).parent / "test.rst"


@dataclass
class FakeConfig:
    no_underscore_emphasis: bool = False
    md_parse_relative_links: bool = False
    md_anonymous_references: bool = False
    md_disable_inline_math: bool = False


@dataclass
class FakeEnv:
    config: FakeConfig = field(default_factory=FakeConfig)


class SmokeTest(unittest.TestCase):
    maxDiff = None

    @unittest.skipIf(
        platform.system() == "Windows",
        "inconsistent column widths on Windows",
    )
    def test_convert(self) -> None:
        content = TEST_MD.read_text()
        expected = TEST_RST.read_text()

        result = convert(content)
        self.assertEqual(expected, result)

    def test_mdinclude_basic(self) -> None:
        content = dedent(
            f"""
            .. mdinclude:: {TEST_MD}

            """
        )
        expected = dedent(
            """\
            <document source="smoke.rst">
                <section ids="title" names="title">
                    <title>
                        Title
                    <section ids="subtitle" names="subtitle">
                        <title>
                            SubTitle
                        <paragraph>
                            <strong>
                                content
            """
        )

        directives.register_directive("mdinclude", MdInclude)
        parser = Parser()
        settings = get_default_settings(Parser)
        settings.env = FakeEnv()
        document = new_document("smoke.rst", settings.copy())
        parser.parse(content, document)

        result = document.pformat()
        self.assertEqual(expected, result[: len(expected)])



from typing import Any, Dict, Match, Tuple

from mistune import BlockParser, InlineParser
from mistune.core import BlockState, InlineState
from mistune.helpers import HTML_ATTRIBUTES, HTML_TAGNAME


State = Dict[str, Any]
Token = Dict[str, Any]
Element = Tuple[str, ...]


class RestBlockParser(BlockParser):
    SPECIFICATION = BlockParser.SPECIFICATION.copy()
    SPECIFICATION.update(
        {
            "directive": r"(?ms:^(?P<directive_1> *\.\..*?)\n(?=\S))",
            "oneline_directive": r"(?ms:^(?P<directive_2> *\.\..*?)$)",
            "rest_code_block": r"(?m:^::\s*$)",
        }
    )

    DEFAULT_RULES = BlockParser.DEFAULT_RULES + (  # type: ignore[has-type]
        "directive",
        "oneline_directive",
        "rest_code_block",
    )

    def parse_directive(self, m: Match[str], state: BlockState) -> int:
        state.append_token({"type": "directive", "raw": m.group("directive_1")})
        return m.end()

    def parse_oneline_directive(self, m: Match[str], state: BlockState) -> int:
        # reuse directive output
        state.append_token({"type": "directive", "raw": m.group("directive_2")})
        # $ does not count '\n'
        return m.end() + 1

    def parse_rest_code_block(self, m: Match[str], state: BlockState) -> int:
        state.append_token({"type": "rest_code_block", "text": ""})
        # $ does not count '\n'
        return m.end() + 1


class RestInlineParser(InlineParser):
    # make inline_html span open/contents/close instead of just a single tag
    INLINE_HTML = (
        r"(?<!\\)<" + HTML_TAGNAME + HTML_ATTRIBUTES + r"\s*>"  # open tag
        r"(.*)"
        r"(?<!\\)</" + HTML_TAGNAME + r"\s*>|"  # close tag
        r"(?<!\\)<" + HTML_TAGNAME + HTML_ATTRIBUTES + r"\s*/>|"  # open/close tag
        r"(?<!\\)<\?[\s\S]+?\?>|"
        r"(?<!\\)<![A-Z][\s\S]+?>|"  # doctype
        r"(?<!\\)<!\[CDATA[\s\S]+?\]\]>"  # cdata
    )

    SPECIFICATION = InlineParser.SPECIFICATION.copy()
    SPECIFICATION.update(
        {
            "inline_html": INLINE_HTML,
            "inline_math": r"`\$(?P<math_1>.*?)\$`",
            "rest_role": r":.*?:`.*?`|`[^`]+`:.*?:",
            "rest_link": r"`[^`]*?`_",
            "eol_literal_marker": r"(?P<eol_space>\s+)?::\s*$",
        }
    )

    # Order is important: need these rules to be checked before the
    # default rules
    DEFAULT_RULES = (
        "inline_math",
        "rest_role",
        "rest_link",
        "eol_literal_marker",
    ) + InlineParser.DEFAULT_RULES  # type: ignore[has-type]

    def parse_rest_role(self, m: Match[str], state: InlineState) -> int:
        """Pass through rest role."""
        state.append_token({"type": "rest_role", "raw": m.group(0)})
        return m.end()

    def parse_rest_link(self, m: Match[str], state: InlineState) -> int:
        """Pass through rest link."""
        state.append_token({"type": "rest_link", "raw": m.group(0)})
        return m.end()

    def parse_inline_math(self, m: Match[str], state: InlineState) -> int:
        """Pass through inline math."""
        state.append_token({"type": "inline_math", "raw": m.group("math_1")})
        return m.end()

    def parse_eol_literal_marker(self, m: Match[str], state: InlineState) -> int:
        """Pass through rest link."""
        marker = ":" if m.group("eol_space") is None else ""
        state.append_token({"type": "eol_literal_marker", "raw": marker})
        # $ does not count '\n'
        return m.end() + 1



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Markdown extension for Sphinx
"""

__author__ = "Hiroyuki Takagi <miyako.dev@gmail.com>"
from .__version__ import __version__

from .render import convert, RestMarkdown
from .sphinx import setup

__all__ = [
    "convert",
    "RestMarkdown",
    "setup",
]



"""
Sphinx extension
"""

import os
import os.path
from typing import Any, Dict, List, Union

from docutils import io, statemachine, utils
from docutils.io import error_string as ErrorString
from docutils.nodes import document as Document
from docutils.parsers import rst
from docutils.parsers.rst import directives as rst_directives
from sphinx.application import Sphinx

from . import RestMarkdown
from .__version__ import __version__


class MdIncludeParser(rst.Parser, object):
    # Explicitly tell supported formats to sphinx
    supported = ("markdown", "md", "mkd")

    def parse(
        self,
        inputstrings: Union[str, statemachine.StringList],
        document: Document,
    ) -> None:
        if isinstance(inputstrings, statemachine.StringList):
            inputstring = "\n".join(inputstrings)
        else:
            inputstring = inputstrings
        config = document.settings.env.config
        converter = RestMarkdown(
            no_underscore_emphasis=config.no_underscore_emphasis,
            parse_relative_links=config.md_parse_relative_links,
            anonymous_references=config.md_anonymous_references,
            disable_inline_math=config.md_disable_inline_math,
        )
        super().parse(converter(inputstring), document)


class MdInclude(rst.Directive):
    """Directive class to include markdown in sphinx.

    Load a file and convert it to rst and insert as a node. Currently
    directive-specific options are not implemented.
    """

    required_arguments = 1
    optional_arguments = 0
    option_spec = {
        "start-line": int,
        "end-line": int,
    }

    def run(self) -> List[Any]:
        """Most of this method is from ``docutils.parser.rst.Directive``.

        docutils version: 0.12
        """
        if not self.state.document.settings.file_insertion_enabled:
            raise self.warning('"%s" directive disabled.' % self.name)
        source = self.state_machine.input_lines.source(
            self.lineno - self.state_machine.input_offset - 1
        )
        source_dir = os.path.dirname(os.path.abspath(source))
        path = rst_directives.path(self.arguments[0])
        path = os.path.normpath(os.path.join(source_dir, path))
        path = utils.relative_path(None, path)
        path = str(path)

        # get options (currently not use directive-specific options)
        encoding = self.options.get(
            "encoding", self.state.document.settings.input_encoding
        )
        e_handler = self.state.document.settings.input_encoding_error_handler
        tab_width = self.options.get(
            "tab-width", self.state.document.settings.tab_width
        )

        # open the including file
        try:
            self.state.document.settings.record_dependencies.add(path)
            include_file = io.FileInput(
                source_path=path, encoding=encoding, error_handler=e_handler
            )
        except UnicodeEncodeError:
            raise self.severe(
                'Problems with "%s" directive path:\n'
                'Cannot encode input file path "%s" '
                "(wrong locale?)." % (self.name, str(path))
            )
        except IOError as error:
            raise self.severe(
                'Problems with "%s" directive path:\n%s.'
                % (self.name, ErrorString(error))
            )

        # read from the file
        startline = self.options.get("start-line", None)
        endline = self.options.get("end-line", None)
        try:
            if startline or (endline is not None):
                lines = include_file.readlines()
                rawtext = "".join(lines[startline:endline])
            else:
                rawtext = include_file.read()
        except UnicodeError as error:
            raise self.severe(
                'Problem with "%s" directive:\n%s' % (self.name, ErrorString(error))
            )

        config = self.state.document.settings.env.config
        converter = RestMarkdown(
            no_underscore_emphasis=config.no_underscore_emphasis,
            parse_relative_links=config.md_parse_relative_links,
            anonymous_references=config.md_anonymous_references,
            disable_inline_math=config.md_disable_inline_math,
        )
        include_lines = statemachine.string2lines(
            converter(rawtext), tab_width, convert_whitespace=True
        )
        self.state_machine.insert_input(include_lines, path)
        return []


def setup(app: Sphinx) -> Dict[str, Union[str, bool]]:
    """When used for sphinx extension."""
    app.add_config_value("no_underscore_emphasis", False, "env")
    app.add_config_value("md_parse_relative_links", False, "env")
    app.add_config_value("md_anonymous_references", False, "env")
    app.add_config_value("md_disable_inline_math", False, "env")
    app.add_source_suffix(".md", "markdown")
    app.add_source_parser(MdIncludeParser)
    app.add_directive("mdinclude", MdInclude)
    metadata: Dict[str, Union[str, bool]] = {
        "version": __version__,
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }
    return metadata
